#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import numpy as np
import pandas
from astropy.io import fits

from ztfquery import query, marshal
from ztfquery.io import LOCALSOURCE

LOCALDATA = LOCALSOURCE + "forcephotometry/"

import os

if os.path.isfile(LOCALSOURCE + "/externaldata/target_position.csv"):
    TARGET_IO = pandas.read_csv(LOCALSOURCE + "/externaldata/target_position.csv")
else:
    TARGET_IO = pandas.DataFrame(
        columns=["name", "mean_ra", "mean_dec", "median_ra", "median_dec"]
    )


class ZTFTarget(object):
    """ """

    # ================ #
    #   Init           #
    # ================ #
    def __init__(self, target=None):
        """ """
        self._zquery = query.ZTFQuery()
        if target is not None:
            self.set_name(target)

    @classmethod
    def from_name(cls, name):
        """ """
        return cls(name)

    @classmethod
    def from_coords(cls, ra, dec, jdmin=None, jdmax=None, name="unknown"):
        """ """
        this = cls()
        this.set_coordinate(ra, dec)
        this.set_jdrange(jdmin, jdmax)
        this.set_name(name)
        return this

    # ================ #
    #  Methods         #
    # ================ #

    # ------- #
    # SETTER  #
    # ------- #
    def set_name(self, targetname):
        """ """
        self._name = targetname

    def set_coordinate(self, ra, dec):
        """ provide the coordinate.
        Get them using self.get_coordinate().
        """
        self._radec = [ra, dec]

    def set_jdrange(self, jdmin, jdmax):
        """ """
        self._jdrange = [jdmin, jdmax]

    # ------- #
    # LOADDER #
    # ------- #
    def load_marshal(self, source="local", program="Cosmology"):
        """ """
        if source == "local":
            self._marshal = marshal.MarshalAccess.load_local(program)
        else:
            self.update_marshal(program=program)

    def update_marshal(self, program="Cosmology", store=True):
        """ """
        self._marshal = marshal.MarshalAccess()
        self._marshal.load_target_sources(program)
        if store:
            self._marshal.store()

    def load_metadata(self, fromname=False):
        """ """
        if self.has_radec() and not fromname:
            # Test if one radec value if given. If not, use 
            # the first one (it does not matter here which
            # one is used)
            if isinstance(self.radec[0], list):
                radec = [self.radec[0][0], self.radec[1][0]]
            else:
                radec = self.radec

            dictmetadata = self.marshal.get_metadataquery(
                *radec, *self.jdrange, size=0.01
            )

        elif self.has_name():
            dictmetadata = self.marshal.get_target_metadataquery(self.name)
        else:
            raise AttributeError("Cannot load the metadata, no self.name and no _radec")

        self.zquery.load_metadata(**dictmetadata)

    def download_data(
        self, which=["scimrefdiffimg.fits.fz", "diffimgpsf.fits"], nprocess=4, **kwargs
    ):
        """

        **kwargs goes to zquery.download_data()
              - source='IRSA',
              - indexes=None,
              - download_dir=None,
              - show_progress=False,
              - notebook=False,
              - verbose=True,
              - nodl=False,
              - overwrite=False,
              - auth=None,
              **kwargs
        """
        for suffix in np.atleast_1d(which):
            self.zquery.download_data(
                suffix, nprocess=nprocess, **{**{"show_progress": False}, **kwargs}
            )

    # ------- #
    # GETTER  #
    # ------- #
    def get_coordinate(self, forcename=False, use="mean"):
        """ """
        if self.has_radec() and not forcename:
            return self.radec

        if len(np.atleast_1d(self.name)) == 1:
            if self.name in TARGET_IO["name"].values:
                return TARGET_IO[TARGET_IO["name"] == self.name][
                    ["%s_ra" % use, "%s_dec" % use]
                ].values[0]

            return self.marshal.get_target_coordinates(self.name).values[0]
        return self.marshal.get_target_coordinates(self.name).values

    def get_jdrange(self, forcetarget=False):
        """ """
        if hasattr(self, "_jdrange") and not forcetarget:
            return self._jdrange

        if len(np.atleast_1d(self.name)) == 1:
            return self.marshal.get_target_jdrange(self.name)
        return self.marshal.get_target_jdrange(self.name)

    # Others
    def get_diffimg_forcepsf_filepath(self, exists=True, indexes=None, **kwargs):
        """
        **kwargs eventually goes to ztfquery.Query.get_local_data()
                 -> filecheck=True, ignore_warnings=False, etc.
        """
        diffimg = self.get_diffimg_filepath(exists=exists, indexes=indexes, **kwargs)
        diffpsf = self.get_diffpsf_filepath(exists=exists, indexes=indexes, **kwargs)
        out = []
        for diff in diffimg:
            expected_psf = diff.replace("scimrefdiffimg.fits.fz", "diffimgpsf.fits")
            if expected_psf not in diffpsf:
                out.append([diff, None])
            else:
                out.append([diff, expected_psf])
        return out

    def get_diffimg_filepath(self, exists=True, indexes=None, **kwargs):
        """
        **kwargs goes to ztfquery.Query.get_local_data()
                 -> filecheck=True, ignore_warnings=False, etc.
        """
        return self.zquery.get_local_data(
            "scimrefdiffimg.fits.fz", exists=exists, indexes=indexes, **kwargs
        )

    def get_diffpsf_filepath(self, exists=True, indexes=None, **kwargs):
        """
        **kwargs goes to ztfquery.Query.get_local_data()
                 -> filecheck=True, ignore_warnings=False, etc.
        """
        return self.zquery.get_local_data(
            "diffimgpsf.fits", exists=exists, indexes=indexes, **kwargs
        )

    # ================ #
    #  Properties      #
    # ================ #
    @property
    def zquery(self):
        """ """
        return self._zquery

    @property
    def marshal(self):
        """ """
        if not hasattr(self, "_marshal"):
            raise AttributeError("Marshal not loaded. See load_marshal()")
        return self._marshal

    @property
    def name(self):
        """ """
        if not self.has_name():
            raise AttributeError("Target not loaded. See set_name()")
        return self._name

    @property
    def jdrange(self):
        """ """
        if not self.has_jdrange():
            return None
        return self._jdrange

    @property
    def radec(self):
        """ """
        if not self.has_radec():
            return None
        return self._radec

    def has_name(self):
        """ Test if the current instance has a target set """
        return hasattr(self, "_name") and self._name is not None

    def has_radec(self):
        """ """
        return hasattr(self, "_radec") and self._radec is not None

    def has_jdrange(self):
        """ """
        return hasattr(self, "_jdrange")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import pandas as pd
import numpy as np
from tqdm import tqdm
import warnings, multiprocessing
from .diffdata import DiffData


def run_forcephotometry(
    target,
    download_files=True,
    nprocess=4,
    verbose=True,
    filecheck=True,
    ignore_warnings=False,
    **kwargs
):
    """ """
    fp = ForcePhotometry.from_name(target)
    fp.load_metadata()
    if download_files:
        fp.io.download_data(nprocess=nprocess)
    fp.load_filepathes(filecheck=filecheck, ignore_warnings=ignore_warnings)
    fp.run_forcefit(update_diffdata=False, verbose=verbose, **kwargs)


class ForcePhotometry:
    """ """

    # =============== #
    #    Init         #
    # =============== #
    def __init__(self, ztftarget):
        """ """
        self.io = ztftarget

    @classmethod
    def from_name(cls, name):
        """Load the ForcePhotometry from Target Name"""
        from .io import ZTFTarget

        return cls(ZTFTarget.from_name(name))

    @classmethod
    def from_coords(cls, ra, dec, jdmin=None, jdmax=None, name="unknown"):
        """Load the ForcePhotometry from Target Name"""
        from .io import ZTFTarget

        return cls(ZTFTarget.from_coords(ra, dec, jdmin=jdmin, jdmax=jdmax, name=name))

    # =============== #
    #   Method        #
    # =============== #
    # -------- #
    # LOADDER  #
    # -------- #
    def load_metadata(self, **kwargs):
        """ """
        if not hasattr(self.io, "marshal"):
            self.io.load_marshal(**kwargs)
        self.io.load_metadata()

    def load_filepathes(self, **kwargs):
        """
        **kwargs eventually goes to ztfquery.Query.get_local_data()
                 -> exists=True, indexes=None, filecheck=True, ignore_warnings=False, etc.

        """
        _filepaths = self.io.get_diffimg_forcepsf_filepath(**kwargs)

        # clean list of missing psf image entries (will be None)
        self._filepathes = [f for f in _filepaths if f[1] is not None]

        self._diffdata = None

    def store(self, filename=None, mode="w"):
        """ """
        if filename is None:
            from .io import LOCALDATA

            filename = LOCALDATA + "/%s.csv" % self.io.name
        self._data_forcefit.to_csv(filename, index=False, mode=mode)

    # -------- #
    # FITTER   #
    # -------- #
    def run_forcefit(
        self,
        indexes=None,
        update_diffdata=False,
        store=False,
        verbose=True,
        no_badsub=False,
        force_refit=False,
        nprocess=4,
    ):
        """ """
        import gc

        if indexes is None:
            indexes = np.arange(self.nsources)

        from .io import LOCALDATA

        path_previous_results = os.path.join(LOCALDATA, "{}.csv".format(self.io.name))

        # Try to load previous fit results
        if os.path.isfile(path_previous_results):
            previous_results = pd.read_csv(
                path_previous_results, float_precision="round_trip", comment="#"
            )
        else:
            previous_results = pd.DataFrame(columns=["filename"])

        dataout = {}
        if verbose:
            print("Starting run_forcefit() for %d image differences" % len(indexes))

        from astropy.utils.console import ProgressBar

        index_count = len(indexes)
        _coords = [self.io.get_coordinate()] * index_count
        _update_diffdata = [update_diffdata] * index_count
        _no_badsub = [no_badsub] * index_count
        _verbose = [verbose] * index_count
        _previous_results = [previous_results] * index_count
        _force_refit = [force_refit] * index_count

        if nprocess > 1:
            bar = ProgressBar(index_count)
            dataout_list = []

            with multiprocessing.Pool(nprocess) as p:
                for j, dataout_i in enumerate(
                    p.imap_unordered(
                        self.get_ith_diffdata_multiprocess,
                        zip(
                            indexes,
                            self.filepathes,
                            _coords,
                            _update_diffdata,
                            _no_badsub,
                            _verbose,
                            _previous_results,
                            _force_refit,
                        ),
                    )
                ):
                    if bar is not None:
                        bar.update(j)
                    dataout_list.append(dataout_i)
                if bar is not None:
                    bar.update(index_count)

            dataout_final = [i for i in dataout_list if i is not None]

            self._data_forcefit = pd.DataFrame(dataout_final)
            if store:
                self.store()

        else:
            for i in tqdm(indexes):
                if verbose:
                    print("running %d " % i)
                    print(self.filepathes[i][0].split("/")[-1])

                filename_modified = self.filepathes[i][0].split("/")[-1][:-26] + ".fits"
                query = previous_results.query("filename == @filename_modified")

                # Check query of dataframe. If it is 40 characters long, this datapoint has already been fitted
                # and will not be fitted again (but it WILL be fitted again if force_refit is True)

                # Parse the query to a dictionary to check if it is 40 characters long or not
                if len(query) == 1:
                    query = query.to_dict("r")[0]

                # Now do the actual check
                if len(query) == 40 and not force_refit:
                    if verbose:
                        print("not fitting %d " % i)
                    dataout[i] = query

                else:
                    if verbose:
                        print("fitting %d " % i)
                    diffdata = self.get_ith_diffdata(i, update=update_diffdata)
                    has_nan = np.any(np.isnan(diffdata.diffimg))
                    if has_nan and no_badsub:
                        print("NaNs in the image, skipped")
                    else:
                        try:
                            with warnings.catch_warnings():
                                warnings.simplefilter("ignore")
                                fitresults = diffdata.fit_flux()
                                datainfo = diffdata.get_main_info()
                                dataout[i] = {**fitresults, **datainfo}
                                dataout[i]["data_hasnan"] = has_nan
                        except ValueError:
                            if verbose:
                                warnings.warn(
                                    "Shape of diffimg and psfimg do not correspond (index: %d). Skipping."
                                    % (i)
                                )
                            pass
                    del diffdata
                gc.collect()

            self._data_forcefit = pd.DataFrame(dataout).T
            if store:
                self.store()

    def get_ith_diffdata(self, index, update=False, rebuild=False, **kwargs):
        """loads and returns a DiffData object corresponding
        at the ith-entry of the self.filepathes

        Parameters
        ----------
        index: [int]
            which entry iof self.filepathes/self.diffdata do you want ?

        update: [bool] -optional-
            if the DiffData is derived (because unknown or rebuild), should this
            be updated to self.diffdata

        rebuild: [bool] -optional-
            Even if self.DiffData[index] exists, shall this remeasure it ?

        Returns
        -------
        DiffData
        """
        if self.diffdata is None or self.diffdata[index] is None or rebuild:
            diffdata = DiffData(
                *self.filepathes[index], self.io.get_coordinate(), **kwargs
            )
            if update:
                self.diffdata[index] = diffdata
            return diffdata

        return self.diffdata[index]

    @staticmethod
    def get_ith_diffdata_multiprocess(args):
        import gc

        (
            index,
            filepath,
            coords,
            update_diffdata,
            no_badsub,
            verbose,
            previous_results,
            force_refit,
        ) = args

        filename_modified = filepath[0].split("/")[-1][:-26] + ".fits"
        query = previous_results.query("filename == @filename_modified")

        # Check query of dataframe. If it is 40 characters long, this datapoint has already been fitted
        # and will not be fitted again (but it WILL be fitted again if force_refit is True)

        # Parse the query to a dictionary to check if it is 40 characters long or not
        if len(query) == 1:
            query = query.to_dict("records")[0]

        # Now do the actual check
        if len(query) == 40 and not force_refit:
            if verbose:
                print("not fitting %d " % index)
            dataout = query

        else:
            if verbose:
                print("fitting %d " % index)
                print(filepath[0].split("/")[-1])
            diffdata = DiffData(*filepath, coords)
            has_nan = np.any(np.isnan(diffdata.diffimg))
            if has_nan and no_badsub:
                if verbose:
                    print("NaNs in the image, skipped")
                return None
            else:
                try:
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        fitresults = diffdata.fit_flux()
                        datainfo = diffdata.get_main_info()
                        dataout = {**fitresults, **datainfo}
                        dataout["data_hasnan"] = has_nan
                except ValueError:
                    if verbose:
                        warnings.warn(
                            "Shape of diffimg and psfimg do not correspond (index: %d). Skipping."
                            % (index)
                        )
                    return None
            del diffdata

        gc.collect()
        return dataout

    # --------- #
    #  PLOTTER  #
    # --------- #
    def show_lc(self, ax=None, scalezp=25, reference=False, **kwargs):
        """ """
        # - Figure
        if ax is None:
            import matplotlib.pyplot as mpl

            fig = mpl.figure()
            ax = fig.add_subplot(111)
        else:
            fig = ax.figure

        x, y, err = self.data_forcefit[["obsmjd", "ampl", "ampl.err"]].values.T
        f0coef = (
            10 ** (-(self.data_forcefit["magzp"] - scalezp) / 2.5)
            if scalezp is not None
            else 1
        )

        for i, band_ in [[1, "C0"], [2, "C2"], [3, "C1"]]:
            if i not in self.data_forcefit["filterid"]:
                continue
            flag = np.asarray(self.data_forcefit["filterid"] == i, dtype="bool")

            prop = {**dict(marker="o", color=band_, zorder=5), **kwargs}
            ax.scatter(x[flag], (y * f0coef)[flag], **prop)
            prop["zorder"] = prop.get("zorder", 5) - 1
            ax.errorbar(
                x[flag],
                (y * f0coef)[flag],
                yerr=err[flag],
                ls="None",
                ecolor="0.6",
                **prop
            )

        ax.axhline(0, ls="--", color="0.5")
        return ax

    # =============== #
    #  Properties     #
    # =============== #
    @property
    def data_forcefit(self):
        """DataFrame containing the forcefit results and data info"""
        if not hasattr(self, "_data_forcefit"):
            raise AttributeError("data_forcefit not loaded. run self.run_forcefit()")
        return self._data_forcefit

    @property
    def nsources(self):
        """ """
        return len(self.filepathes)

    @property
    def metadata(self):
        """ """
        return self.io.zquery.metatable

    @property
    def filepathes(self):
        """List of DiffImage and PSFDiffImage file path"""
        if not hasattr(self, "_filepathes"):
            self.load_filepathes()
        return self._filepathes

    def had_filepathes(self):
        """ """
        return hasattr(self, "_filepathes") and self._filepathes is not None

    @property
    def diffdata(self):
        """DiffData object associated with the filepathes"""
        if not hasattr(self, "_diffdata") or self._diffdata is None:
            if self.had_filepathes():
                self._diffdata = [None for i in range(self.nsources)]
            else:
                self._diffdata = None

        return self._diffdata



#! /usr/bin/env python
# -*- coding: utf-8 -*-


def make_method(obj):
    """Decorator to make the function a method of *obj*.

    In the current context::
      @make_method(Axes)
      def toto(ax, ...):
          ...
    makes *toto* a method of `Axes`, so that one can directly use::
      ax.toto()
    COPYRIGHT: from Yannick Copin
    """

    def decorate(f):
        setattr(obj, f.__name__, f)
        return f

    return decorate



#! /usr/bin/env python
# -*- coding: utf-8 -*-


#################################
#
#   MAIN
#
#################################
if __name__ == "__main__":

    import argparse
    import numpy as np
    from ztflc import forcephotometry

    # ================= #
    #   Options         #
    # ================= #
    parser = argparse.ArgumentParser(
        description=""" run the interactive plotting of a given cube""",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument("infile", type=str, default="None", help="Target Name")

    parser.add_argument(
        "--download",
        action="store_true",
        default=False,
        help="Download the data if necessary ?",
    )

    parser.add_argument(
        "--ndlprocess",
        type=int,
        default=4,
        help="Number of parallel processing for the download",
    )

    # // Other stuffs
    parser.add_argument(
        "--quiet", action="store_true", default=False, help="Set the verbose to False"
    )

    # Parsing
    args = parser.parse_args()

    # ================= #
    #   Script          #
    # ================= #
    forcephotometry.run_forcephotometry(
        args.infile,
        download_files=args.download,
        nprocess=args.ndlprocess,
        verbose=~args.quiet,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import numpy as np
from scipy import stats
import warnings
from iminuit import Minuit

from .utils import make_method


class DiffImgFitter(object):

    FREEPARAMETERS = ["sigma", "ampl"]

    def __new__(cls, *arg, **kwargs):
        """ """
        obj = super(DiffImgFitter, cls).__new__(cls)
        exec(
            "@make_method(DiffImgFitter)\n"
            + "def _minuit_chi2_(self,%s): \n" % (", ".join(obj.FREEPARAMETERS))
            + "    parameters = %s \n" % (", ".join(obj.FREEPARAMETERS))
            + "    return self.get_chi2(parameters)\n"
        )
        return obj

    def __init__(self, diffimg, psfimg, diffvar=None, shape=None):
        """ """
        self.set_data(diffimg, psfimg, diffvar=diffvar, shape=shape)

    # ============== #
    #   FITTER       #
    # ============== #
    def fit(self, verbose=False, **kwargs):
        """ """
        self._hfitverbose = verbose
        # Setting guesses
        self._setup_minuit_(**kwargs)
        # fitting
        self.minuit.migrad()

        # checking fit
        if not self.minuit.valid:
            if self._hfitverbose:
                warnings.warn("migrad is not valid")
            self.fit_ok = False
        else:
            self.fit_ok = True

        # saving results
        self.best_parameters = np.asarray(
            [self.minuit.values[k] for k in self.FREEPARAMETERS]
        )
        # - readout -
        self.fitvalues = {}
        covmat = self.get_fit_covmatrix()

        for i, name in enumerate(self.FREEPARAMETERS):
            self.fitvalues[name] = self.best_parameters[i]
            self.fitvalues[name + ".err"] = np.sqrt(covmat[i, i])

        # -- Additional data -- #
        self.fitvalues["fval"] = self.minuit.fval
        self.fitvalues["chi2"] = self.get_chi2(simple_chi2=True)
        self.fitvalues["chi2dof"] = self.fitvalues["chi2"] / self.dof

        return self.fitvalues

    def _setup_minuit_(self, step=1, print_level=0, **kwargs):
        """ """
        # == Minuit Keys == #
        minuit_kwargs = {}
        for param in self.FREEPARAMETERS:
            minuit_kwargs[param] = kwargs.get(
                "%s_guess" % param, 0 if not "sigma" in param else 1
            )

        self._minuit_input = minuit_kwargs

        self.minuit = Minuit(self._minuit_chi2_, **minuit_kwargs)
        self.minuit.errordef = step
        self.minuit.print_level = print_level

        for param in self.FREEPARAMETERS:
            limits = kwargs.get(
                "%s_boundaries" % param, None if not "sigma" in param else [0, None]
            )
            if limits:
                self.minuit.limits[param] = (limits[0], limits[1])
            fixed = kwargs.get("%s_fixed" % param, False)
            self.minuit.fixed[param] = fixed

    # -------- #
    # SETTER   #
    # -------- #
    def set_data(self, diffimg, psfimg, diffvar=None, shape=None):
        """ """
        if np.shape(diffimg) != np.shape(psfimg):
            raise ValueError(
                "diffimg and psfimg do not have the same shape. They must."
            )

        if diffvar is not None:
            if len(np.atleast_1d(diffvar)) == 1:
                self.diffvar = float(np.atleast_1d(diffvar)[0])
            elif np.shape(diffimg) == np.shape(diffvar):
                self.diffvar = np.asarray(diffvar)
            else:
                raise ValueError(
                    "diffvar is not a single value and diffimg and diffvar do not have the same shape. "
                )
        else:
            self.diffvar = 0

        self._inputshape = np.shape(diffimg)
        self.data = np.ravel(diffimg)
        self.flagout = np.isnan(self.data)
        self.profile = np.ravel(psfimg)
        self._shape = shape

    def set_parameters(self, param):
        """ """
        self.parameters = {k: v for k, v in zip(self.FREEPARAMETERS, param)}

    # ============== #
    #   GETTER       #
    # ============== #
    def get_chi2(self, param=None, simple_chi2=False, use_prior=True):
        """ """
        if param is not None:
            self.set_parameters(param)

        if self._fitverbose:
            print(self.parameters)
        if simple_chi2:
            return np.nansum((self.data - self.scaled_model) ** 2 / self.variance_tot)

        if not use_prior:
            return -2 * np.nansum(self.get_loglikelihood())
        return -2 * np.nansum(self.get_logproba())

    # // Posterior
    def get_logproba(self):
        """ """
        val = self.get_loglikelihood() + self.get_logprior()
        if self._fitverbose:
            print("get_logproba:", val)
        return val

    # // Likelihood
    def get_loglikelihood(self):
        """ """
        val = stats.norm.logpdf(
            self.data, loc=self.scaled_model, scale=np.sqrt(self.variance_tot)
        )
        if self._fitverbose:
            print("get_loglikelihood:", val)
        return val

    # // Priors
    def get_logprior(self):
        """ """
        return 0

    def get_fit_covmatrix(self):
        """coveriance matrix after the fit"""

        def _read_hess_(hess):
            """ """
            if len(hess) == len(self.FREEPARAMETERS):
                return hess

            indexFixed = [
                i
                for i, name in enumerate(self.FREEPARAMETERS)
                if "%s_fixed" % name in self._minuit_input
            ]
            for i in indexFixed:
                newhess = np.insert(hess, i, 0, axis=0)
                newhess = np.insert(newhess, i, 0, axis=1)
                hess = newhess

            return hess

        if self.minuit.valid:
            return _read_hess_(np.asarray(self.minuit.covariance))
        else:
            fakeMatrix = np.zeros(
                (len(self.best_parameters), len(self.best_parameters))
            )
            for i, k in enumerate(self.FREEPARAMETERS):
                fakeMatrix[i, i] = self.minuit.errors[k] ** 2
            if self._hfitverbose:
                warnings.warn("Inaccurate covariance Matrix. Only trace defined")
            return _read_hess_(fakeMatrix)

    # --------- #
    #  PLOTTER  #
    # --------- #
    def show(self, axes=None, res_aspull=True):
        """ """
        if axes is not None:
            axd, axm, axr = axes
            fig = axd.figure
        else:
            import matplotlib.pyplot as mpl

            fig = mpl.figure(figsize=[9, 3])
            axd = fig.add_axes([0.08, 0.1, 0.25, 0.8])
            axm = fig.add_axes([0.38, 0.1, 0.25, 0.8])
            axr = fig.add_axes([0.68, 0.1, 0.25, 0.8])

        data_ = self.data
        axd.imshow(data_.reshape(self._shape), origin="lower")

        model_ = self.scaled_model
        axm.imshow(model_.reshape(self._shape), origin="lower")
        # Residual
        if res_aspull:
            norm = np.sqrt(self.variance_tot)
            prop = dict(vmin=-3, vmax=3)
        else:
            norm = 1
            prop = {}
        axr.imshow(
            ((data_ - model_) / norm).reshape(self._shape), origin="lower", **prop
        )
        if res_aspull:
            axr.text(
                0.05,
                0.95,
                "Pull [-3,3]",
                color="w",
                va="top",
                ha="left",
                transform=axr.transAxes,
                fontdict=dict(weight="bold"),
            )
        return fig

    # ============== #
    #  Properties    #
    # ============== #
    @property
    def variance_tot(self):
        """ """
        return self.diffvar + self.parameters["sigma"] ** 2

    @property
    def scaled_model(self):
        """ """
        return self.profile * self.parameters["ampl"]

    @property
    def ndata(self):
        """ """
        return len(self.data)

    @property
    def nfreeparameters(self):
        """ """
        fixed = (
            0
            if not hasattr(self, "_minuit_input")
            else len([k for k, v in self._minuit_input.items() if "fixed" in k and v])
        )
        return len(self.FREEPARAMETERS) - fixed

    @property
    def dof(self):
        """ """
        return self.ndata - self.nfreeparameters

    @property
    def _fitverbose(self):
        """ """
        if not hasattr(self, "_hfitverbose"):
            self._hfitverbose = False
        return self._hfitverbose



#! /usr/bin/env python
#

DESCRIPTION = "ztflc: Force photometry lc fitter"
LONG_DESCRIPTION = """ Force photometry lc fitter"""

DISTNAME = "ztflc"
AUTHOR = "Mickael Rigault"
MAINTAINER = "Mickael Rigault"
MAINTAINER_EMAIL = "m.rigault@ipnl.in2p3.fr"
URL = "https://github.com/MickaelRigault/ztflc/"
LICENSE = "BSD (3-clause)"
DOWNLOAD_URL = "https://github.com/MickaelRigault/ztflc/tarball/0.2"
VERSION = "0.2.8"

try:
    from setuptools import setup, find_packages

    _has_setuptools = True
except ImportError:
    from distutils.core import setup

    _has_setuptools = False


def check_dependencies():
    install_requires = []

    # Just make sure dependencies exist, I haven't rigorously
    # tested what the minimal versions that will work are
    # (help on that would be awesome)
    try:
        import ztfquery
    except ImportError:
        install_requires.append("ztfquery")

    try:
        import pandas
    except ImportError:
        install_requires.append("pandas")

    try:
        import iminuit
    except ImportError:
        install_requires.append("iminuit")

    try:
        import numpy
    except ImportError:
        install_requires.append("numpy")

    return install_requires


if __name__ == "__main__":

    install_requires = check_dependencies()

    if _has_setuptools:
        packages = find_packages()
        print(packages)
    else:
        # This should be updated if new submodules are added
        packages = ["ztflc"]

    setup(
        name=DISTNAME,
        author=AUTHOR,
        author_email=MAINTAINER_EMAIL,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        license=LICENSE,
        url=URL,
        version=VERSION,
        download_url=DOWNLOAD_URL,
        install_requires=install_requires,
        scripts=["bin/forcephoto.py"],
        packages=packages,
        include_package_data=True,
        # package_data={'pysedm': ['data/*.*']},
        classifiers=[
            "Intended Audience :: Science/Research",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3.5",
            "License :: OSI Approved :: BSD License",
            "Topic :: Scientific/Engineering :: Astronomy",
            "Operating System :: POSIX",
            "Operating System :: Unix",
            "Operating System :: MacOS",
        ],
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import warnings
import numpy as np
from astropy.io import fits
from scipy import stats


class DiffData(object):
    """ """

    def __init__(self, diffimgpath, psfimgpath, coords, inpixels=False, clean=True):
        """ """
        self.set_data(diffimgpath, psfimgpath, coords, inpixels=inpixels, clean=clean)

    # ================ #
    #    Methods       #
    # ================ #
    # ------- #
    # FITTER  #
    # ------- #
    def fit_flux(self, **kwargs):
        """ """
        from .fitter import DiffImgFitter
        from astropy.stats import mad_std

        # Load the fitter
        self.fitter = DiffImgFitter(self.diffimg, self.psfimg, 0, shape=self.psfshape)
        robust_nmad = mad_std(self.fitter.data[~np.isnan(self.fitter.data)])
        fit_prop = {
            "ampl_guess": np.nanmax(self.fitter.data)
            * (np.sqrt(2 * np.pi * 2**2)),  # 2 sigma guess
            "sigma_guess": robust_nmad,
        }
        fit_prop["ampl_boundaires"] = [
            -2 * np.nanmin(self.fitter.data) * (np.sqrt(2 * np.pi * 2**2)),
            5 * np.nanmax(self.fitter.data) * (np.sqrt(2 * np.pi * 2**2)),
        ]
        fit_prop["sigma_boundaries"] = [
            robust_nmad / 10.0,
            np.nanstd(self.fitter.data) * 2,
        ]

        # Return fitter output
        return self.fitter.fit(**{**fit_prop, **kwargs})

    # ------- #
    # SETTER  #
    # ------- #
    def set_data(
        self, diffimgpath, psfimgpath, coords, inpixels=False, clean=True, **kwargs
    ):
        """ """
        from astropy import wcs
        from astropy.io import fits
        from scipy import ndimage
        from astropy.stats import mad_std

        self._xy = coords if inpixels else None
        #
        # Handeling too many open files
        #
        # PSF
        with fits.open(psfimgpath) as psf:
            self._psfimg_raw = psf[0].data.copy()

        with fits.open(diffimgpath) as fdiff:
            # If coords is a list, then assume different coordinates for each
            # filter
            filterid = fdiff[1].header["FILTERID"]
            imgshape = fdiff[1].data.shape

            if isinstance(coords[0], list):
                ra = coords[0][filterid - 1]
                dec = coords[1][filterid - 1]
                coords = [ra, dec]
            # x,y position
            if self._xy is None:
                wcs_ = wcs.WCS(fdiff[1].header)
                self._xy = wcs_.all_world2pix([coords], 0)[0]
                del wcs_

            # data position
            x, y = np.asarray(self._xy, dtype="int")
            buffer = (np.asarray(self.psfshape)) / 2 - 0.5
            xmin, xmax = int(x - buffer[0]), int(x + buffer[0] + 1)
            ymin, ymax = int(y - buffer[1]), int(y + buffer[1] + 1)
            self._diffimg = fdiff[1].data[ymin:ymax, xmin:xmax].copy()
            self._datatmp = fdiff[1].data.copy()
            self._diffimg_targetpos = self._xy - [xmin, ymin]
            self._header = fdiff[1].header.copy()
            self._istarget_in = (
                (x - buffer[0] > 0)
                and (x + buffer[0] < imgshape[1])
                and (y - buffer[1] > 0)
                and (y + buffer[1] < imgshape[0])
            )

        if clean:
            self._iscleaned = True
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                flagout = self._diffimg < -mad_std(self._diffimg) * 10
            self._diffimg[flagout] = np.NaN
        else:
            self._iscleaned = False

        self._psf_shift = self._diffimg_targetpos - np.asarray(self.psfshape) / 2 + 0.5
        # ::-1 because numpy arrays are not images but matrices
        self._psfimg = ndimage.interpolation.shift(
            self._psfimg_raw, self._psf_shift[::-1], order=5
        )

    # ------- #
    # GETTER  #
    # ------- #
    def get_main_info(self, backup_value=None):
        """ """
        header = {
            k.lower(): self.header.get(k) if k in self.header else backup_value
            for k in self._main_columns
        }
        header["target_x"], header["target_y"] = self.target_position
        return header

    def estimate_noise(self, buffer=10, use="nmad"):
        """ """
        raise NotImplementedError("estimate_noise is not ready yet")
        """
        from shapely import geometry
        from astropy.stats import mad_std

        circle = geometry.Point(*self.target_position).buffer(buffer)
        pixelsout, flagbuffer = self.grid_diffimg.get_pixels_in(circle, invert=True)

        dataout = self.grid_diffimg.geodataframe[flagbuffer]["data"]
        self.noise_data = {"flagnoise":flagbuffer,
                            "mean": np.mean(dataout),
                            "std": np.std(dataout),
                            "median": np.median(dataout),
                            "nmad":mad_std(dataout),
                            "buffer":buffer,
                            }

        self.noise_data["mean.err"] = self.noise_data["nmad"]/np.sqrt(len(dataout)-1)
        self.noise_data["p(zero)"]  = stats.norm.pdf(self.noise_data["mean"], loc=0, scale=self.noise_data["mean.err"])
        self.noise_data["differr"]  = self.noise_data[use]
        return self.noise_data["differr"]
        """

    # ------- #
    # PLOTTER #
    # ------- #
    def show(self):
        """ """
        import matplotlib.pyplot as mpl

        fig = mpl.figure(figsize=[8, 3])
        prop = dict(origin="lower")  # vmin=-20,vmax=60)

        ax = fig.add_subplot(131)
        ax.imshow(self.diffimg, **prop)

        ax = fig.add_subplot(132)
        ax.imshow(self.psfimg_raw, **prop)

        ax = fig.add_subplot(133)
        ax.imshow(self.psfimg, **prop)

    def show_alignment(self, ax=None):
        """ """
        # - Data
        ax = self.grid_diffimg.show("data", ax=ax, lw=0)
        ax.scatter(*self.target_position, marker="x", color="C1", s=80)
        # - Grid out
        _ = self.grid_psf.show(None, ax=ax, lw=0.5)

    def show_noise(self, ax=None):
        """ """
        import matplotlib.pyplot as mpl
        from matplotlib.patches import Circle

        fig = mpl.figure(figsize=[9, 3])
        axleft = fig.add_axes([0.05, 0.12, 0.3, 0.8], frameon=False)
        ax = fig.add_axes([0.4, 0.12, 0.45, 0.78])

        dataout = self.grid_diffimg.geodataframe[self.noise_data["flagnoise"]]["data"]
        datain = self.grid_diffimg.geodataframe[~self.noise_data["flagnoise"]]["data"]

        _ = self.grid_diffimg.show("data", ax=axleft, lw=0)
        axleft.add_patch(
            Circle(
                self.target_position,
                self.noise_data["buffer"],
                facecolor="None",
                edgecolor="C1",
                lw=2,
            )
        )
        axleft.scatter(*self.target_position, marker="x", color="C1", s=80)

        range = [
            int(self.noise_data["median"] - 5 * self.noise_data["nmad"]),
            int(self.noise_data["median"] + 5 * self.noise_data["nmad"]),
        ]
        ax.hist(
            dataout,
            density=True,
            range=range,
            bins=15,
            facecolor="C0",
            histtype="step",
            fill=True,
            alpha=0.4,
            label="out data",
        )
        ax.hist(
            dataout,
            density=True,
            range=range,
            bins=15,
            edgecolor="C0",
            histtype="step",
            lw=2,
        )
        ax.hist(
            datain,
            histtype="step",
            density=True,
            range=[range[0], range[1] * 2],
            bins=20,
            color="C1",
            lw=2,
            label="in data",
        )

        mean, std = np.mean(dataout), np.std(dataout)
        xx = np.linspace(*range, 100)
        ax.plot(xx, stats.norm.pdf(xx, mean, std), ls="--", color="C0")
        ax.axvline(
            mean,
            label=r"$%.2f \pm %.2f$"
            % (self.noise_data["mean"], self.noise_data["mean.err"]),
            color="C0",
        )

        fig.text(
            0.01,
            0.98,
            "Estimated error: %.2f" % self.estimated_noise,
            va="top",
            ha="left",
        )

        ax.legend()

    # ================ #
    #  Properties      #
    # ================ #
    #
    # Data Image
    #
    @property
    def diffimg(self):
        """Difference data patch around the target."""
        if not hasattr(self, "_diffimg"):
            self._diffimg = None
        return self._diffimg

    @property
    def psfimg_raw(self):
        """PSF data as given by the IRSA pipeline"""
        if not hasattr(self, "_psfimg_raw"):
            self._psfimg_raw = None
        return self._psfimg_raw

    @property
    def psfimg(self):
        """PSF data re-aligned with the target"""
        if not hasattr(self, "_psfimg_raw"):
            self._psfimg = None
        return self._psfimg

    @property
    def psfshape(self):
        """shape of the psf image"""
        return np.shape(self.psfimg_raw)

    #
    # target
    #
    @property
    def target_position(self):
        """ """
        if not hasattr(self, "_xy"):
            self._xy = None
        return self._xy

    @property
    def target_in(self):
        """Test if the target is enough within the image"""
        if not hasattr(self, "_istarget_in"):
            return None
        return self._istarget_in

    #
    # Additional information
    #
    @property
    def header(self):
        """Data header"""
        if not hasattr(self, "_header"):
            self._header = None
        return self._header

    @property
    def _main_columns(self):
        """ """
        return [
            "FILENAME",
            "HUMIDITY",
            "FILTER",
            "OBSMJD",
            "CCDID",
            "AMP_ID",
            "GAIN",
            "READNOI",
            "DARKCUR",
            "MAGZP",
            "MAGZPUNC",
            "MAGZPRMS",
            "CLRCOEFF",
            "CLRCOUNC",
            "ZPCLRCOV",
            "ZPMED",
            "ZPAVG",
            "ZPRMSALL",
            "CLRMED",
            "CLRAVG",
            "CLRRMS",
            "QID",
            "RCID",
            "SEEING",
            "AIRMASS",
            "NMATCHES",
            "MAGLIM",
            "STATUS",
            "INFOBITS",
            "FILTERID",
            "FILTER",
            "FIELDID",
            "MOONALT",
            "MOONILLF",
        ]
